/**
 * @license
 * Copyright 2019 Google LLC. All Rights Reserved.
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * =============================================================================
 */
/// <reference types="@webgpu/types" />
import './flags_webgpu';
import { backend_util, DataMover, DataType, KernelBackend, Rank, ShapeMap, Tensor, Tensor3D, Tensor4D, TimingInfo } from '@tensorflow/tfjs-core';
import * as shaderc from '@webgpu/shaderc';
import { BufferManager } from './buffer_manager';
declare type MemoryInfo = {
    numTensors: number;
    numDataBuffers: number;
    numBytes: number;
    unreliable?: boolean;
    reasons: string[];
};
interface RecursiveArray<T extends any> {
    [index: number]: T | RecursiveArray<T>;
}
export interface WebGPUMemoryInfo extends MemoryInfo {
    numBytesInGPU: number;
    unreliable: boolean;
}
interface DataId {
}
export interface CPUTimerQuery {
    startMs: number;
    endMs: number;
}
export declare type WebGPUKernelInfo = {
    name: string;
    query: Promise<number>;
};
export declare type TimerNode = RecursiveArray<WebGPUKernelInfo> | WebGPUKernelInfo;
export interface WebGPUTimingInfo extends TimingInfo {
    uploadWaitMs: number;
    downloadWaitMs: number;
}
export declare class WebGPUBackend extends KernelBackend {
    device: GPUDevice;
    queue: GPUQueue;
    shaderc: shaderc.Shaderc;
    compiler: shaderc.Compiler;
    compileOpts: shaderc.CompileOptions;
    commandQueue: GPUCommandEncoder[];
    private commandQueueOwnedIds;
    private binaryCache;
    private fromPixels2DContext;
    private bufferManager;
    private tensorMap;
    private disposalQueue;
    private disposed;
    private programTimersStack;
    private activeTimers;
    private uploadWaitMs;
    private downloadWaitMs;
    constructor(device: GPUDevice, shaderc: shaderc.Shaderc);
    floatPrecision(): 32;
    setDataMover(dataMover: DataMover): void;
    flushDisposalQueue(): void;
    disposeData(dataId: DataId): void;
    memory(): WebGPUMemoryInfo;
    getBufferManager(): BufferManager;
    private acquireBuffer;
    private releaseBuffer;
    register(dataId: object, shape: number[], dtype: DataType): void;
    write(dataId: object, values: Float32Array | Int32Array | Uint8Array): void;
    private submitQueue;
    private getBufferData;
    private convertAndCacheOnCPU;
    readSync(dataId: object): Float32Array | Int32Array | Uint8Array;
    read(dataId: object): Promise<Float32Array | Int32Array | Uint8Array>;
    time(f: () => void): Promise<WebGPUTimingInfo>;
    private getAndSavePipeline;
    private makeOutputArray;
    private tensorToBinding;
    startTimer(): {
        startMs: number;
        endMs: number;
    };
    endTimer(query: CPUTimerQuery): CPUTimerQuery;
    getQueryTime(query: CPUTimerQuery): Promise<number>;
    private compileAndRun;
    private makeUniforms;
    pad<T extends Tensor>(x: T, paddings: Array<[number, number]>, constantValue: number): T;
    maxPool(x: Tensor4D, convInfo: backend_util.Conv2DInfo): Tensor4D;
    private binaryOp;
    add(a: Tensor, b: Tensor): Tensor;
    subtract(a: Tensor, b: Tensor): Tensor;
    conv2d(x: Tensor4D, filter: Tensor4D, convInfo: backend_util.Conv2DInfo): Tensor4D;
    depthwiseConv2D(x: Tensor4D, filter: Tensor4D, convInfo: backend_util.Conv2DInfo): Tensor4D;
    private argMinMaxReduce;
    argMin(x: Tensor, axis: number): Tensor;
    argMax(x: Tensor, axis: number): Tensor;
    concat(tensors: Tensor[], axis: number): Tensor;
    multiply(a: Tensor, b: Tensor): Tensor;
    realDivide(a: Tensor, b: Tensor): Tensor;
    floorDiv(a: Tensor, b: Tensor): Tensor;
    sigmoid<T extends Tensor>(x: T): T;
    relu<T extends Tensor>(x: T): T;
    resizeBilinear(x: Tensor4D, newHeight: number, newWidth: number, alignCorners: boolean): Tensor4D;
    reshape<R extends Rank>(x: Tensor, shape: ShapeMap[R]): Tensor<R>;
    cast<T extends Tensor>(x: T, dtype: DataType): T;
    transpose<T extends Tensor>(x: T, perm: number[]): T;
    batchMatMul(a: Tensor3D, b: Tensor3D, transposeA: boolean, transposeB: boolean): Tensor3D;
    fromPixels(pixels: backend_util.PixelData | ImageData | HTMLImageElement | HTMLCanvasElement | HTMLVideoElement, numChannels: number): Tensor3D;
    dispose(): void;
}
export {};
