"use strict";
/**
 * @license
 * Copyright 2019 Google LLC. All Rights Reserved.
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * =============================================================================
 */
Object.defineProperty(exports, "__esModule", { value: true });
const shader_preprocessor = require("../shader_preprocessor");
exports.makeBindGroup = (device, bindGroupLayout, inputs, output, uniforms) => {
    const bindings = [output, ...inputs];
    if (uniforms) {
        bindings.push(uniforms);
    }
    return device.createBindGroup({
        layout: bindGroupLayout,
        bindings: bindings.map((b, i) => ({ binding: i, resource: b.resource })),
    });
};
const makeBindGroupLayout = (device, inputs, output, uniforms) => {
    const bindings = Array(1 + inputs.length).fill({
        visibility: GPUShaderStage.COMPUTE,
        type: 'storage-buffer'
    });
    if (uniforms) {
        bindings.push({
            visibility: GPUShaderStage.COMPUTE,
            type: 'uniform-buffer'
        });
    }
    return device.createBindGroupLayout({
        bindings: bindings.map((b, i) => (Object.assign({ binding: i }, b))),
    });
};
exports.compileProgram = (shaderCompiler, shaderKind, compileOptions, device, program, inputsData, output, uniforms) => {
    const outputData = { dtype: output.dtype, shape: output.shape };
    const source = shader_preprocessor.makeShader(inputsData, outputData, program);
    const result = shaderCompiler.CompileGlslToSpv(source, shaderKind, 'file', 'main', compileOptions);
    const error = result.GetErrorMessage();
    if (error.length) {
        console.error(source.split('\n')
            .map((s, l) => (l + 1).toString().padStart(5, ' ') + ' ' + s)
            .join('\n'));
        throw new Error(`Shader compilation failed: ${error}`);
    }
    const bindGroupLayout = makeBindGroupLayout(device, inputsData, output, uniforms);
    const code = result.GetBinary();
    const layout = device.createPipelineLayout({ bindGroupLayouts: [bindGroupLayout] });
    const module = device.createShaderModule({ code });
    const pipeline = device.createComputePipeline({ layout, computeStage: { module, entryPoint: 'main' } });
    return { bindGroupLayout, pipeline };
};
// TODO: Consider allowing each program to specify its own shader key. E.g. some
// kernels account for different work group sizes, but some don't.
// TODO: Consider uploading shape info as vec4s regardless of rank to reduce
// recompilation.
function makeShaderKey(program, ranks) {
    const key = (program.workGroupSize ? program.workGroupSize.join(',') : '') +
        ranks.join(',') + program.userCode;
    return key;
}
exports.makeShaderKey = makeShaderKey;
//# sourceMappingURL=webgpu_program.js.map